@@ -2,7 +2,7 @@
 '''
 
 TSCODE: Transition State Conformational Docker
-Copyright (C) 2021 Nicolò Tampellini
+Copyright (C) 2021-2022 Nicolò Tampellini
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -21,24 +21,27 @@ import time
 import warnings
 from copy import deepcopy
 
+import matplotlib.pyplot as plt
 import numpy as np
 from ase import Atoms
-from ase.calculators.calculator import PropertyNotImplementedError
+from ase.calculators.calculator import (CalculationFailed,
+                                        PropertyNotImplementedError)
 from ase.calculators.gaussian import Gaussian
 from ase.calculators.mopac import MOPAC
 from ase.calculators.orca import ORCA
 from ase.constraints import FixInternals
 from ase.dyneb import DyNEB
 from ase.optimize import BFGS, LBFGS
+from ase.vibrations import Vibrations
 from rmsd import kabsch
 from sella import Sella
 
-from tscode.fast_algebra import norm, norm_of
-from tscode.hypermolecule_class import graphize
+from tscode.algebra import norm, norm_of
+from tscode.graph_manipulations import findPaths, graphize, neighbors
 from tscode.settings import COMMANDS, MEM_GB
 from tscode.solvents import get_solvent_line
-from tscode.utils import (HiddenPrints, clean_directory, findPaths,
-                          get_double_bonds_indexes, molecule_check, neighbors,
+from tscode.utils import (HiddenPrints, clean_directory,
+                          get_double_bonds_indexes, molecule_check,
                           scramble_check, time_to_string, write_xyz)
 
 
@@ -49,11 +52,11 @@ class Spring:
     Spring constant is very high to achieve tight convergence,
     but force is dampened so as not to ruin structures.
     '''
-    def __init__(self, i1, i2, d_eq, k=1000, tight=False):
+    def __init__(self, i1, i2, d_eq, k=100, tight=False):
         self.i1, self.i2 = i1, i2
         self.d_eq = d_eq
         self.k = k
-        self.tight = False
+        self.tight = tight
 
     def adjust_positions(self, atoms, newpositions):
         pass
@@ -76,6 +79,7 @@ class Spring:
 
     def tighten(self):
         self.tight = True
+        self.k = 1000
 
     def __repr__(self):
         return f'Spring - ids:{self.i1}/{self.i2} - d_eq:{self.d_eq}, k:{self.k}'
@@ -115,22 +119,22 @@ class HalfSpring:
     def __repr__(self):
         return f'Spring - ids:{self.i1}/{self.i2} - d_max:{self.d_max}, k:{self.k}'
 
-def get_ase_calc(docker):
+def get_ase_calc(embedder):
     '''
     Attach the correct ASE calculator
     to the ASE Atoms object.
-    docker: either a TSCoDe docker object or
+    embedder: either a TSCoDe embedder object or
         a 4-element strings tuple containing
         (calculator, method, procs, solvent)
     '''
-    if isinstance(docker, tuple):
-        calculator, method, procs, solvent = docker
+    if isinstance(embedder, tuple):
+        calculator, method, procs, solvent = embedder
 
     else:
-        calculator = docker.options.calculator
-        method = docker.options.theory_level
-        procs = docker.options.procs
-        solvent = docker.options.solvent
+        calculator = embedder.options.calculator
+        method = embedder.options.theory_level
+        procs = embedder.options.procs
+        solvent = embedder.options.solvent
 
     if calculator == 'XTB':
         try:
@@ -142,6 +146,7 @@ def get_ase_calc(docker):
 
         from tscode.solvents import (solvent_synonyms, xtb_solvents,
                                      xtb_supported)
+                                     
         solvent = solvent_synonyms[solvent] if solvent in solvent_synonyms else solvent
         solvent = 'none' if solvent is None else solvent
 
@@ -207,36 +212,36 @@ def get_ase_calc(docker):
 
             return calc
 
-def ase_adjust_spacings(docker, structure, atomnos, constrained_indexes, title=0, traj=None):
+def ase_adjust_spacings(embedder, structure, atomnos, constrained_indexes, title=0, traj=None):
     '''
-    docker: TSCoDe docker object
+    embedder: TSCoDe embedder object
     structure: TS candidate coordinates to be adjusted
     atomnos: 1-d array with element numbering for the TS
     constrained_indexes: (n,2)-shaped array of indexes to be distance constrained
     mols_graphs: list of NetworkX graphs, ordered as the single molecules in the TS
-    title: number to be used for referring to this structure in the docker log
+    title: number to be used for referring to this structure in the embedder log
     traj: if set to a string, traj+'.traj' is used as a filename for the refinement trajectory.
     '''
     atoms = Atoms(atomnos, positions=structure)
 
-    atoms.calc = get_ase_calc(docker)
+    atoms.calc = get_ase_calc(embedder)
     
-    springs = [Spring(indexes[0], indexes[1], dist) for indexes, dist in docker.target_distances.items()]
+    springs = [Spring(indexes[0], indexes[1], dist) for indexes, dist in embedder.target_distances.items()]
     # adding springs to adjust the pairings for which we have target distances
 
-    nci_indexes = [indexes for letter, indexes in docker.pairings_table.items() if letter in ('x', 'y', 'z')]
+    nci_indexes = [indexes for letter, indexes in embedder.pairings_table.items() if letter in ('x', 'y', 'z')]
     halfsprings = [HalfSpring(i1, i2, 2.5) for i1, i2 in nci_indexes]
     # HalfSprings get atoms involved in NCIs together if they are more than 2.5A apart,
     # but lets them achieve their natural equilibrium distance when closer
 
     psc = PreventScramblingConstraint(graphize(structure, atomnos),
                                         atoms,
-                                        double_bond_protection=docker.options.double_bond_protection,
-                                        fix_angles=docker.options.fix_angles_in_deformation)
+                                        double_bond_protection=embedder.options.double_bond_protection,
+                                        fix_angles=embedder.options.fix_angles_in_deformation)
 
     atoms.set_constraint(springs + halfsprings + [psc])
 
-    t_start_opt = time.time()
+    t_start_opt = time.perf_counter()
     try:
         with LBFGS(atoms, maxstep=0.2, logfile=None, trajectory=traj) as opt:
 
@@ -258,28 +263,30 @@ def ase_adjust_spacings(docker, structure, atomnos, constrained_indexes, title=0
 
         new_structure = atoms.get_positions()
 
-        success = scramble_check(new_structure, atomnos, constrained_indexes, docker.graphs)
+        success = scramble_check(new_structure, atomnos, constrained_indexes, embedder.graphs)
         exit_str = 'REFINED' if success else 'SCRAMBLED'
 
     except PropertyNotImplementedError:
         exit_str = 'CRASHED'
 
-    docker.log(f'    - {docker.options.calculator} {docker.options.theory_level} refinement: Structure {title} {exit_str} ({iterations} iterations, {time_to_string(time.time()-t_start_opt)})', p=False)
+    embedder.log(f'    - {title} {exit_str} ({iterations} iterations, {time_to_string(time.perf_counter()-t_start_opt)})', p=False)
 
     if exit_str == 'CRASHED':
         return None, None, False
 
-    return new_structure, atoms.get_total_energy(), success
+    energy = atoms.get_total_energy() * 23.06054194532933 #eV to kcal/mol
 
-def ase_saddle(docker, coords, atomnos, constrained_indexes=None, mols_graphs=None, title='temp', logfile=None, traj=None, freq=False, maxiterations=200):
+    return new_structure, energy, success
+
+def ase_saddle(embedder, coords, atomnos, constrained_indexes=None, mols_graphs=None, title='temp', logfile=None, traj=None, freq=False, maxiterations=200):
     '''
     Runs a first order saddle optimization through the ASE package
     '''
     atoms = Atoms(atomnos, positions=coords)
 
-    atoms.calc = get_ase_calc(docker)
+    atoms.calc = get_ase_calc(embedder)
     
-    t_start = time.time()
+    t_start = time.perf_counter()
     with HiddenPrints():
         with Sella(atoms,
                    logfile=None,
@@ -290,7 +297,7 @@ def ase_saddle(docker, coords, atomnos, constrained_indexes=None, mols_graphs=No
             iterations = opt.nsteps
 
     if logfile is not None:
-        t_end_berny = time.time()
+        t_end_berny = time.perf_counter()
         elapsed = t_end_berny - t_start
         exit_str = 'converged' if iterations < maxiterations else 'stopped'
         logfile.write(f'{title} - {exit_str} in {iterations} steps ({time_to_string(elapsed)})\n')
@@ -298,28 +305,49 @@ def ase_saddle(docker, coords, atomnos, constrained_indexes=None, mols_graphs=No
     new_structure = atoms.get_positions()
     energy = atoms.get_total_energy() * 23.06054194532933 #eV to kcal/mol
 
-    # if freq:
-    #     vib = Vibrations(atoms, name='temp')
-    #     with HiddenPrints():
-    #         vib.run()
-    #     freqs = vib.get_frequencies()
-
-    #     if logfile is not None:
-    #         elapsed = time.time() - t_end_berny
-    #         logfile.write(f'{title} - frequency calculation completed ({time_to_string(elapsed)})\n')
-        
-    #     return new_structure, energy, freqs
-
     if mols_graphs is not None:
-        success = scramble_check(new_structure, atomnos, constrained_indexes, mols_graphs, max_newbonds=docker.options.max_newbonds)
+        success = scramble_check(new_structure, atomnos, constrained_indexes, mols_graphs, max_newbonds=embedder.options.max_newbonds)
     else:
-        success = molecule_check(coords, new_structure, atomnos, max_newbonds=docker.options.max_newbonds)
+        success = molecule_check(coords, new_structure, atomnos, max_newbonds=embedder.options.max_newbonds)
 
     return new_structure, energy, success
 
-def ase_neb(docker, reagents, products, atomnos, n_images=6, title='temp', optimizer=LBFGS, logfile=None):
+def ase_vib(embedder, coords, atomnos, logfunction=None, title='temp'):
     '''
-    docker: tscode docker object
+    '''
+    atoms = Atoms(atomnos, positions=coords)
+    atoms.calc = get_ase_calc(embedder)
+    vib = Vibrations(atoms, name=title)
+
+    if os.path.isdir(title):
+        os.chdir(title)
+        for f in os.listdir():
+            os.remove(f)
+        os.chdir(os.path.dirname(os.getcwd()))
+    else:
+        os.mkdir(title)
+
+    os.chdir(title)
+
+    t_start = time.perf_counter()
+
+    with HiddenPrints():
+        vib.run()
+
+    # freqs = vib.get_frequencies()
+    freqs = vib.get_energies()* 8065.544 # from eV to cm-1 
+
+    if logfunction is not None:
+        elapsed = time.perf_counter() - t_start
+        logfunction(f'{title} - frequency calculation completed ({time_to_string(elapsed)})')
+    
+    os.chdir(os.path.dirname(os.getcwd()))
+
+    return freqs, np.count_nonzero(freqs.imag > 1e-3)
+
+def ase_neb(embedder, reagents, products, atomnos, n_images=6, title='temp', optimizer=LBFGS, logfunction=None, write_plot=False, verbose_print=False):
+    '''
+    embedder: tscode embedder object
     reagents: coordinates for the atom arrangement to be used as reagents
     products: coordinates for the atom arrangement to be used as products
     atomnos: 1-d array of atomic numbers
@@ -343,12 +371,15 @@ def ase_neb(docker, reagents, products, atomnos, n_images=6, title='temp', optim
     neb.interpolate()
 
     ase_dump(f'{title}_MEP_guess.xyz', images, atomnos)
+
+    if verbose_print and logfunction is not None:
+        logfunction(f'\n\n--> Saved interpolated MEP guess to {title}_MEP_guess.xyz\n')
     
     # Set calculators for all images
     for _, image in enumerate(images):
-        image.calc = get_ase_calc(docker)
+        image.calc = get_ase_calc(embedder)
 
-    t_start = time.time()
+    t_start = time.perf_counter()
 
     # Set the optimizer and optimize
     try:
@@ -358,7 +389,7 @@ def ase_neb(docker, reagents, products, atomnos, n_images=6, title='temp', optim
             # ignore runtime warnings from the NEB module:
             # if something went wrong, we will deal with it later
 
-            with optimizer(neb, maxstep=0.1, logfile=None) as opt:
+            with optimizer(neb, maxstep=0.1, logfile=None if not verbose_print else '-') as opt:
 
                 opt.run(fmax=0.05, steps=20)
                 # some free relaxation before starting to climb
@@ -367,17 +398,18 @@ def ase_neb(docker, reagents, products, atomnos, n_images=6, title='temp', optim
                 opt.run(fmax=0.05, steps=500)
                 iterations = opt.nsteps
 
-    except PropertyNotImplementedError:
-        if logfile is not None:
-            logfile.write(f'    - NEB for {title} CRASHED ({time_to_string(time.time()-t_start)})\n')
+    except (PropertyNotImplementedError, CalculationFailed):
+        if logfunction is not None:
+            logfunction(f'    - NEB for {title} CRASHED ({time_to_string(time.perf_counter()-t_start)})\n')
         return None, None, False
 
     exit_status = 'CONVERGED' if iterations < 499 else 'MAX ITER'
 
-    if logfile is not None:
-        logfile.write(f'    - NEB for {title} {exit_status} ({time_to_string(time.time()-t_start)})\n')
+    if logfunction is not None:
+        logfunction(f'    - NEB for {title} {exit_status} ({time_to_string(time.perf_counter()-t_start)})\n')
 
-    energies = [image.get_total_energy() for image in images]
+    energies = [image.get_total_energy() * 23.06054194532933 for image in images]
+    # eV to kcal/mol
     ts_id = energies.index(max(energies))
     # print(f'TS structure is number {ts_id}, energy is {max(energies)}')
 
@@ -385,7 +417,33 @@ def ase_neb(docker, reagents, products, atomnos, n_images=6, title='temp', optim
     ase_dump(f'{title}_MEP.xyz', images, atomnos)
     # Save the converged MEP (minimum energy path) to an .xyz file
 
-    return images[ts_id].get_positions(), images[ts_id].get_total_energy(), True
+    if write_plot:
+
+        plt.figure()
+        plt.plot(
+            range(1,len(images)+1),
+            np.array(energies)-min(energies),
+            color='tab:blue',
+            label='Image energies',
+            linewidth=3,
+        )
+
+        plt.plot(
+            [ts_id+1],
+            [energies[ts_id]-min(energies)],
+            color='gold',
+            label='TS guess',
+            marker='o',
+            markersize=3,
+        )
+
+        plt.legend()
+        plt.title(title)
+        plt.xlabel(f'Image number')
+        plt.ylabel('Rel. E. (kcal/mol)')
+        plt.savefig(f'{title.replace(" ", "_")}_plt.svg')
+
+    return images[ts_id].get_positions(), energies[ts_id], True
 
 class OrbitalSpring:
     '''
@@ -495,59 +553,61 @@ def PreventScramblingConstraint(graph, atoms, double_bond_protection=False, fix_
 
     return FixInternals(dihedrals_deg=dihedrals_deg, angles_deg=angles_deg, bonds=bonds, epsilon=1)
 
-def ase_popt(docker, coords, atomnos, constrained_indexes=None, targets=None, safe=False, safe_mask=None, traj=None):
+def ase_popt(embedder, coords, atomnos, constrained_indexes=None,
+             steps=500, targets=None, safe=False, safe_mask=None,
+             traj=None, logfunction=None, title='temp'):
     '''
-    docker: TSCoDe docker object
+    embedder: TSCoDe embedder object
     coords: 
     atomnos: 
     constrained_indexes:
     safe: if True, adds a potential that prevents atoms from scrambling
     safe_mask: bool array, with False for atoms to be excluded when calculating bonds to preserve
-    title: name to be used for referring to this structure in the docker log
-    traj: if set to a string, traj+'.traj' is used as a filename for the bending trajectory.
+    traj: if set to a string, traj is used as a filename for the bending trajectory.
           not only the atoms will be printed, but also all the orbitals and the active pivot.
-    check: if True, after bending checks that the bent structure did not scramble.
-           If it did, returns the initial molecule.
     '''
     atoms = Atoms(atomnos, positions=coords)
-    atoms.calc = get_ase_calc(docker)
+    atoms.calc = get_ase_calc(embedder)
     constraints = []
 
-    for i, c in enumerate(constrained_indexes):
-        i1, i2 = c
-        tgt_dist = norm_of(coords[i1]-coords[i2]) if targets is None else targets[i]
-        constraints.append(Spring(i1, i2, tgt_dist))
+    if constrained_indexes is not None:
+        for i, c in enumerate(constrained_indexes):
+            i1, i2 = c
+            tgt_dist = norm_of(coords[i1]-coords[i2]) if targets is None else targets[i]
+            constraints.append(Spring(i1, i2, tgt_dist))
 
     if safe:
         constraints.append(PreventScramblingConstraint(graphize(coords, atomnos, safe_mask),
                                                         atoms,
-                                                        double_bond_protection=docker.options.double_bond_protection,
-                                                        fix_angles=docker.options.fix_angles_in_deformation))
+                                                        double_bond_protection=embedder.options.double_bond_protection,
+                                                        fix_angles=embedder.options.fix_angles_in_deformation))
 
     atoms.set_constraint(constraints)
 
-    # t_start_opt = time.time()
-    with LBFGS(atoms, maxstep=0.2, logfile=None, trajectory=traj) as opt:
-        opt.run(fmax=0.05, steps=500)
+    t_start_opt = time.perf_counter()
+    with LBFGS(atoms, maxstep=0.1, logfile=None, trajectory=traj) as opt:
+        opt.run(fmax=0.05, steps=steps)
         iterations = opt.nsteps
 
     new_structure = atoms.get_positions()
     success = (iterations < 499)
 
-    # exit_str = 'REFINED' if success else 'MAX ITER'
+    if logfunction is not None:
+        exit_str = 'REFINED' if success else 'MAX ITER'
+        logfunction(f'    - {title} {exit_str} ({iterations} iterations, {time_to_string(time.perf_counter()-t_start_opt)})')
 
-    # docker.log(f'    - {docker.options.calculator} {docker.options.theory_level} POPT: Structure {title} {exit_str} ({iterations} iterations, {time_to_string(time.time()-t_start_opt)})', p=False)
+    energy = atoms.get_total_energy() * 23.06054194532933 #eV to kcal/mol
 
-    return new_structure, atoms.get_total_energy(), success
+    return new_structure, energy, success
 
-def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=None, check=True):
+def ase_bend(embedder, original_mol, conf, pivot, threshold, title='temp', traj=None, check=True):
     '''
-    docker: TSCoDe docker object
+    embedder: TSCoDe embedder object
     original_mol: Hypermolecule object to be bent
     conf: index of conformation in original_mol to be used
     pivot: pivot connecting two Hypermolecule orbitals to be approached/distanced
     threshold: target distance for the specified pivot, in Angstroms
-    title: name to be used for referring to this structure in the docker log
+    title: name to be used for referring to this structure in the embedder log
     traj: if set to a string, traj+'.traj' is used as a filename for the bending trajectory.
           not only the atoms will be printed, but also all the orbitals and the active pivot.
     check: if True, after bending checks that the bent structure did not scramble.
@@ -556,8 +616,8 @@ def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=No
 
     identifier = np.sum(original_mol.atomcoords[conf])
 
-    if hasattr(docker, "ase_bent_mols_dict"):
-        cached = docker.ase_bent_mols_dict.get((identifier, tuple(sorted(pivot.index)), round(threshold, 3)))
+    if hasattr(embedder, "ase_bent_mols_dict"):
+        cached = embedder.ase_bent_mols_dict.get((identifier, tuple(sorted(pivot.index)), round(threshold, 3)))
         if cached is not None:
             return cached
 
@@ -602,7 +662,7 @@ def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=No
 
     atoms = Atoms(mol.atomnos, positions=mol.atomcoords[conf])
 
-    atoms.calc = get_ase_calc(docker)
+    atoms.calc = get_ase_calc(embedder)
     
     if traj is not None:
         traj_obj = Trajectory(traj + f'_conf{conf}.traj',
@@ -614,7 +674,7 @@ def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=No
 
     unproductive_iterations = 0
     break_reason = 'MAX ITER'
-    t_start = time.time()
+    t_start = time.perf_counter()
 
     for iteration in range(500):
 
@@ -628,8 +688,8 @@ def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=No
 
         c2 = PreventScramblingConstraint(mol.graph,
                                             atoms,
-                                            double_bond_protection=docker.options.double_bond_protection,
-                                            fix_angles=docker.options.fix_angles_in_deformation)
+                                            double_bond_protection=embedder.options.double_bond_protection,
+                                            fix_angles=embedder.options.fix_angles_in_deformation)
 
         atoms.set_constraint([
                                 c1,
@@ -667,7 +727,7 @@ def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=No
             atom.init(mol, index, update=True, orb_dim=orb_memo[index])
             # orbitals positions are calculated based on the conformer we are working on
 
-        temp_pivots = docker._get_pivots(mol)[0]
+        temp_pivots = embedder._get_pivots(mol)[0]
 
         for p in temp_pivots:
             if p.index == pivot.index:
@@ -684,14 +744,14 @@ def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=No
         # else:
             # print('delta is ', round(dist - threshold, 3))
 
-        docker.log(f'    {title} - conformer {conf} - {break_reason}{" "*(9-len(break_reason))} ({iteration+1}{" "*(3-len(str(iteration+1)))} iterations, {time_to_string(time.time()-t_start)})', p=False)
+    embedder.log(f'    {title} - conformer {conf} - {break_reason}{" "*(9-len(break_reason))} ({iteration+1}{" "*(3-len(str(iteration+1)))} iterations, {time_to_string(time.perf_counter()-t_start)})', p=False)
 
-        if check:
-            if not molecule_check(original_mol.atomcoords[conf], mol.atomcoords[0], mol.atomnos, max_newbonds=1):
-                mol.atomcoords[0] = original_mol.atomcoords[conf]
-            # keep the bent structures only if no scrambling occurred between atoms
+    if check:
+        if not molecule_check(original_mol.atomcoords[conf], mol.atomcoords[0], mol.atomnos, max_newbonds=1):
+            mol.atomcoords[0] = original_mol.atomcoords[conf]
+        # keep the bent structures only if no scrambling occurred between atoms
 
-        final_mol.atomcoords[conf] = mol.atomcoords[0]
+    final_mol.atomcoords[conf] = mol.atomcoords[0]
 
     # Now align the ensembles on the new reactive atoms positions
 
@@ -715,11 +775,11 @@ def ase_bend(docker, original_mol, conf, pivot, threshold, title='temp', traj=No
         for index, atom in final_mol.reactive_atoms_classes_dict[conf_].items():
             atom.init(final_mol, index, update=True, orb_dim=orb_memo[index])
 
-    docker._set_pivots(final_mol)
+    embedder._set_pivots(final_mol)
 
     # add result to cache (if we have it) so we avoid recomputing it
-    if hasattr(docker, "ase_bent_mols_dict"):
-        docker.ase_bent_mols_dict[(identifier, tuple(sorted(pivot.index)), round(threshold, 3))] = final_mol
+    if hasattr(embedder, "ase_bent_mols_dict"):
+        embedder.ase_bent_mols_dict[(identifier, tuple(sorted(pivot.index)), round(threshold, 3))] = final_mol
 
     clean_directory()
 
@@ -729,4 +789,4 @@ def ase_dump(filename, images, atomnos):
     with open(filename, 'w') as f:
                 for i, image in enumerate(images):
                     coords = image.get_positions()
-                    write_xyz(coords, atomnos, f, title=f'{filename[:-4]}_image_{i}')
+                    write_xyz(coords, atomnos, f, title=f'{filename[:-4]}_image_{i}')
\ No newline at end of file
