@@ -2,7 +2,7 @@
 '''
 
 TSCODE: Transition State Conformational Docker
-Copyright (C) 2021 Nicolò Tampellini
+Copyright (C) 2021-2022 Nicolò Tampellini
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -25,13 +25,14 @@ from ase.optimize import LBFGS
 from networkx.algorithms.components.connected import connected_components
 from networkx.algorithms.shortest_paths.generic import shortest_path
 
+from tscode.algebra import dihedral
 from tscode.ase_manipulations import ase_neb, ase_saddle, get_ase_calc
 from tscode.hypermolecule_class import align_structures, graphize
-from tscode.utils import (clean_directory, dihedral, loadbar, molecule_check,
-                   time_to_string, write_xyz)
+from tscode.utils import (clean_directory, loadbar, molecule_check,
+                          time_to_string, write_xyz)
 
 
-def ase_torsion_TSs(docker,
+def ase_torsion_TSs(embedder,
                     coords,
                     atomnos,
                     indexes,
@@ -74,8 +75,13 @@ def ase_torsion_TSs(docker,
                 logfile.write(s+'\n')
 
     else:
+
+        if not embedder.options.let:
+            raise SystemExit('The specified dihedral angle is made up of non-contiguous atoms. To prevent errors, the\n' +
+                             'run has been stopped. Override this behavior with the LET keyword.')
+
         # if user did not provide four contiguous indexes,
-        # just move the fourth atom and
+        # and did that on purpose, just move the fourth atom and
         # let the rest of the structure relax
         indexes_to_be_moved = [i4]
         # cyclical = True
@@ -96,7 +102,7 @@ def ase_torsion_TSs(docker,
         if logfile is not None:
             logfile.write('\n')
 
-        structures, energies = ase_scan(docker,
+        structures, energies = ase_dih_scan(embedder,
                                         coords,
                                         atomnos,
                                         indexes=indexes,
@@ -109,11 +115,13 @@ def ase_torsion_TSs(docker,
                                         logfile=logfile)
 
         min_e = min(energies)
-        output_structures, output_energies = [], []
+        rel_energies = [e-min_e for e in energies]
+
+        tag = '_relaxed' if optimization else '_rigid'
         
-        for structure, energy in zip(structures, energies):
-            output_structures.append(structure)
-            output_energies.append(energy)
+        with open(title + tag + direction + '_scan.xyz', 'w') as outfile:
+            for s, structure in enumerate(align_structures(np.array(structures), indexes[:-1])):
+                write_xyz(structure, atomnos, outfile, title=f'Scan point {s+1}/{len(structures)} - Rel. E = {round(rel_energies[s], 3)} kcal/mol')
 
         if plot:
             import pickle
@@ -122,18 +130,18 @@ def ase_torsion_TSs(docker,
 
             fig = plt.figure()
 
-            x1 = [dihedral(structures[i][indexes]) for i in range(steps)]
+            x1 = [dihedral(structure[indexes]) for structure in structures]
             y1 = [e-min_e for e in energies]
 
-            x1, y1 = zip(*sorted(zip(x1, y1), key=lambda x: x[0]))
+            for i, (x_, y_) in enumerate(get_plot_segments(x1, y1, max_step=abs(degrees)+1)):
 
-            plt.plot(x1,
-                    y1,
-                    '-',
-                    color='tab:blue',
-                    label='Preliminary SCAN'+direction,
-                    linewidth=3,
-                    alpha=0.50)
+                plt.plot(x_,
+                        y_,
+                        '-',
+                        color='tab:blue',
+                        label=('Preliminary SCAN'+direction) if i == 0 else None,
+                        linewidth=3,
+                        alpha=0.50)
 
         peaks_indexes = atropisomer_peaks(energies, min_thr=min_e+threshold_kcal, max_thr=min_e+75)
 
@@ -147,7 +155,7 @@ def ase_torsion_TSs(docker,
 
             for p, peak in enumerate(peaks_indexes):
 
-                sub_structures, sub_energies = ase_scan(docker,
+                sub_structures, sub_energies = ase_dih_scan(embedder,
                                                         structures[peak-1],
                                                         atomnos,
                                                         indexes=indexes,
@@ -162,36 +170,38 @@ def ase_torsion_TSs(docker,
                 if logfile is not None:
                     logfile.write('\n')
 
-                # for sub_structure, sub_energy in zip(sub_structures, sub_energies):
-                #     output_structures.append(sub_structure)
-                #     output_energies.append(sub_energy)
-
-                # This was a tentative to include accurate scan structures into the
-                # final scan file, but yields rough results and i do not like it
-
                 if plot:
                     x2 = [dihedral(structure[indexes]) for structure in sub_structures]
                     y2 = [e-min_e for e in sub_energies]
 
-                    x2, y2 = zip(*sorted(zip(x2, y2), key=lambda x: x[0]))
+                    for i, (x_, y_) in enumerate(get_plot_segments(x2, y2, max_step=abs(degrees/10)+1)):
 
-                    plt.plot(x2, 
-                            y2,
-                            '-o',
-                            color='tab:red',
-                            label='Accurate SCAN' if p == 0 else None,
-                            markersize=1,
-                            linewidth=2,
-                            alpha=0.5)
+                        plt.plot(x_, 
+                                y_,
+                                '-o',
+                                color='tab:red',
+                                label='Accurate SCAN' if (p == 0 and i == 0) else None,
+                                markersize=1,
+                                linewidth=2,
+                                alpha=0.5)
 
                 sub_peaks_indexes = atropisomer_peaks(sub_energies, min_thr=threshold_kcal+min_e, max_thr=min_e+75)
 
                 if sub_peaks_indexes:
 
                     s = 's' if len(sub_peaks_indexes) > 1 else ''
-                    s = f'Found {len(sub_peaks_indexes)} sub-peak{s}.' + (
-                        f'Performing Saddle opt optimization{s}.' if docker.options.saddle else '')
-                    print(s)
+                    msg = f'Found {len(sub_peaks_indexes)} sub-peak{s}.'
+                    
+                    if embedder.options.saddle or embedder.options.neb:
+                        if embedder.options.saddle:
+                            tag = 'saddle'
+                        else:
+                            tag = 'NEB TS'
+
+                        msg += f'Performing {tag} optimization{s}.'
+
+                    print(msg)
+
                     if logfile is not None:
                         logfile.write(s+'\n')
 
@@ -202,13 +212,13 @@ def ase_torsion_TSs(docker,
                             y = sub_energies[sub_peak]-min_e
                             plt.plot(x, y, color='gold', marker='o', label='Maxima' if p == 0 else None, markersize=3)
 
-                        if docker.options.saddle:
+                        if embedder.options.saddle:
 
                             loadbar_title = f'  > Saddle opt on sub-peak {s+1}/{len(sub_peaks_indexes)}'
                             # loadbar(s+1, len(sub_peaks_indexes), loadbar_title+' '*(29-len(loadbar_title)))
                             print(loadbar_title)
                         
-                            optimized_geom, energy, _ = ase_saddle(docker,
+                            optimized_geom, energy, _ = ase_saddle(embedder,
                                                                     sub_structures[sub_peak],
                                                                     atomnos,
                                                                     title=f'Saddle opt - peak {p+1}, sub-peak {s+1}',
@@ -219,25 +229,24 @@ def ase_torsion_TSs(docker,
                                 ts_structures.append(optimized_geom)
                                 energies.append(energy)
 
-                        elif docker.options.neb:
+                        elif embedder.options.neb:
 
-                            for s, sub_peak in enumerate(sub_peaks_indexes):
-
-                                loadbar_title = f'  > NEB TS opt on sub-peak {s+1}/{len(sub_peaks_indexes)}'
-                                # loadbar(s+1, len(sub_peaks_indexes), loadbar_title+' '*(29-len(loadbar_title)))
-                                print(loadbar_title)
+                            loadbar_title = f'  > NEB TS opt on sub-peak {s+1}/{len(sub_peaks_indexes)}, {direction[1:]}'
+                            drctn = 'clkws' if direction == '_clockwise' else 'ccws'
                             
-                                optimized_geom, energy, success = ase_neb(docker,
-                                                                            sub_structures[sub_peak-2],
-                                                                            sub_structures[sub_peak+1],
-                                                                            atomnos,
-                                                                            n_images=5,
-                                                                            title=f'NEB_peak_{p+1}_sub-peak_{s+1}',
-                                                                            logfile=logfile)
+                            print(loadbar_title)
+                        
+                            optimized_geom, energy, success = ase_neb(embedder,
+                                                                        sub_structures[sub_peak-2],
+                                                                        sub_structures[(sub_peak+1)%len(sub_structures)],
+                                                                        atomnos,
+                                                                        n_images=5,
+                                                                        title=f'{title}_NEB_peak_{p+1}_sub-peak_{s+1}_{drctn}',
+                                                                        logfunction=embedder.log)
 
-                                if success and molecule_check(coords, optimized_geom, atomnos):
-                                    ts_structures.append(optimized_geom)
-                                    energies.append(energy)
+                            if success and molecule_check(coords, optimized_geom, atomnos):
+                                ts_structures.append(optimized_geom)
+                                energies.append(energy)
 
                         else:
                             ts_structures.append(sub_structures[sub_peak])
@@ -254,18 +263,6 @@ def ase_torsion_TSs(docker,
             if logfile is not None:
                 logfile.write('No suitable peaks found.\n\n')
 
-        # output_structures, output_energies = zip(*sorted(zip(output_structures, output_energies), key=lambda x: dihedral(x[0][indexes])))
-        # Sorting structures and energies based on scanned dihedral
-        # angle value, to get a continuous scan in the .xyz file
-
-        rel_energies = [e-min_e for e in output_energies]
-
-        tag = '_relaxed' if optimization else '_rigid'
-
-        with open(title + tag + direction + '_scan.xyz', 'w') as outfile:
-            for s, structure in enumerate(align_structures(np.array(output_structures), indexes[:-1])):
-                write_xyz(structure, atomnos, outfile, title=f'Scan point {s+1}/{len(output_structures)} - Rel. E = {round(rel_energies[s], 3)} kcal/mol')
-
         if plot:
             plt.legend()
             plt.xlabel(f'Dihedral Angle {tuple(indexes)}')
@@ -282,20 +279,26 @@ def ase_torsion_TSs(docker,
 def atropisomer_peaks(data, min_thr, max_thr):
     '''
     data: iterable
-    threshold: peaks must be greater than threshold
-    cap: peaks must be less than cap
+    min_thr: peaks must be values greater than min_thr
+    max_thr: peaks must be values smaller than max_thr
     return: list of peak indexes
     '''
     l = len(data)
-    peaks = [i for i in range(l) if (
+    peaks = [i for i in range(l-1) if (
 
-        data[i-1] < data[i] >= data[(i+1)%l] and
-        max_thr > data[i] > min_thr # discard peaks that are too small or too big
+        data[i-1] < data[i] >= data[i+1] and
+        # peaks have neighbors that are smaller than them
+
+        max_thr > data[i] > min_thr and
+        # discard peaks that are too small or too big
+
+        abs(data[i] - min((data[i-1], data[i+1]))) > 2
+        # discard small peaks (noise) that do not "drop" at least 2 kcal/mol
     )]
 
     return peaks
     
-def ase_scan(docker,
+def ase_dih_scan(embedder,
             coords,
             atomnos,
             indexes,
@@ -307,6 +310,7 @@ def ase_scan(docker,
             title='temp scan',
             logfile=None):
     '''
+    Performs a dihedral scan via the ASE library
     if ad libitum, steps is the minimum number of performed steps
     '''
     assert len(indexes) == 4
@@ -318,7 +322,7 @@ def ase_scan(docker,
     atoms = Atoms(atomnos, positions=coords)
     structures, energies = [], []
 
-    atoms.calc = get_ase_calc(docker)
+    atoms.calc = get_ase_calc(embedder)
 
     if indexes_to_be_moved is None:
         indexes_to_be_moved = range(len(atomnos))
@@ -398,3 +402,22 @@ def ase_scan(docker,
         logfile.write(f'{title} - completed ({time_to_string(elapsed)})\n')
 
     return align_structures(structures, indexes), energies
+
+def get_plot_segments(x, y, max_step=2):
+    '''
+    Returns a zip object with x, y segments.
+    A single segment has x values with separation
+    smaller than max_step.
+    '''
+    x, y = zip(*sorted(zip(x, y), key=lambda t: t[0]))
+    
+    x_slices, y_slices = [], []
+    for i, n in enumerate(x):
+        if abs(x[i-1]-n) > max_step:
+            x_slices.append([])
+            y_slices.append([])
+
+        x_slices[-1].append(n)
+        y_slices[-1].append(y[i])
+
+    return zip(x_slices, y_slices)
