@@ -2,7 +2,7 @@
 '''
 
 TSCODE: Transition State Conformational Docker
-Copyright (C) 2021 Nicolò Tampellini
+Copyright (C) 2021-2022 Nicolò Tampellini
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -42,4 +42,9 @@ class CCReadError(Exception):
 class MopacReadError(Exception):
     '''
     Thrown when reading MOPAC output files fails for some reason.
+    '''
+
+class SegmentedGraphError(Exception):
+    '''
+    Thrown by Clustered CSearch when graph has more than one connected component.
     '''
\ No newline at end of file
