@@ -2,7 +2,7 @@
 '''
 
 TSCODE: Transition State Conformational Docker
-Copyright (C) 2021 Nicolò Tampellini
+Copyright (C) 2021-2022 Nicolò Tampellini
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -16,13 +16,10 @@ GNU General Public License for more details.
 
 '''
 
-from settings import CALCULATOR, FF_CALC, FF_OPT_BOOL, MEM_GB, PROCS
-
-
 def run_setup():
     '''
     Invoked by the command
-    > python -m tscode -setup
+    > python -m tscode -s (--setup)
 
     Guides the user in setting up the calculation options
     contained in the settings.py file.
@@ -71,7 +68,11 @@ def run_setup():
     answer = ask('What Force Field calculator would you like to use?\n- XTB -> xtb\n- Openbabel -> ob\n- Gaussian -> gau\n'
                  '- None -> none\n\nAnswer [xtb]/ob/gau/none: ', accepted=('xtb', 'ob', 'gau', 'none'), default='xtb')
 
-    if answer != 'none':
+    if answer == 'xtb':
+        properties['FF_OPT_BOOL'] = True
+        properties['FF_CALC'] = 'XTB'
+
+    elif answer != 'none':
         properties['FF_OPT_BOOL'] = True
         properties['FF_CALC'] = tag_dict[answer]
 
@@ -82,7 +83,7 @@ def run_setup():
 
     #########################################################################################
 
-    answer = ask('What calculator would like to use?\n- MOPAC -> mop\n- ORCA -> orca\n- Gaussian -> gau\n- XTB -> xtb\n\nAnswer mop/orca/gau/[xtb]: ',
+    answer = ask('What main calculator would you like to use?\n- MOPAC -> mop\n- ORCA -> orca\n- Gaussian -> gau\n- XTB -> xtb\n\nAnswer mop/orca/gau/[xtb]: ',
                 accepted=('mop','orca','gau','xtb'), default='xtb')
 
     properties['CALCULATOR'] = tag_dict[answer]
@@ -104,9 +105,9 @@ def run_setup():
 
     #########################################################################################
 
-    if properties['CALCULATOR'] in ('ORCA', 'GAUSSIAN'):
-        properties['PROCS'] = ask(f'How many cores should {properties["CALCULATOR"]} jobs run on? [1] : ',
-                                    accepted=[str(n) for n in range(1,1000)], default=1)
+    
+    properties['PROCS'] = ask(f'How many cores should {properties["CALCULATOR"]} jobs run on? [{os.cpu_count()}] : ',
+                                accepted=[str(n) for n in range(1,1000)], default=os.cpu_count())
 
     #########################################################################################
 
@@ -174,12 +175,12 @@ def run_setup():
 
     ff = f'{FF_CALC}/{DEFAULT_FF_LEVELS[FF_CALC]}' if FF_OPT_BOOL else 'Turned off'
     opt = f'{CALCULATOR}/{DEFAULT_LEVELS[CALCULATOR]}'
-    s = f'  FF    : {ff}\n  OPT   : {opt}'
+    s = f'  FF    : {ff}\n  OPT   : {opt}\n PROCS  :{PROCS}'
 
-    if CALCULATOR in ('GAUSSIAN', 'ORCA'):
-        s += f'\n  PROCS : {PROCS}'
+    if CALCULATOR == 'GAUSSIAN':
+        s += f'\n  MEM   : {MEM_GB} GB'
 
-        if CALCULATOR == 'GAUSSIAN':
-            s += f'\n  MEM   : {MEM_GB} GB'
+    print(s)
 
-    print(s)
\ No newline at end of file
+if __name__ == '__main__':
+    run_setup()
\ No newline at end of file
