@@ -2,7 +2,7 @@
 '''
 
 TSCODE: Transition State Conformational Docker
-Copyright (C) 2021 Nicolò Tampellini
+Copyright (C) 2021-2022 Nicolò Tampellini
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -19,16 +19,16 @@ GNU General Public License for more details.
 # IF YOU MANUALLY EDIT THIS FILE, BE SURE NOT TO
 # CHANGE IDENTATION/WHITESPACES/NEWLINES!
 
-FF_OPT_BOOL = False
+FF_OPT_BOOL = True
 # Whether to run Force Field optimization with
 # prior to the final one. Set to False if no
 # Openbabel/XTB programs and python bindings
 # are installed.
 
-FF_CALC = 'None'
+FF_CALC = 'OB'
 # Calculator to perform Force Field optimizations.
 # Possibilites are:
-# 'OB' : Openbabel UFF and MMFF methods
+# 'OB' : Openbabel UFF and MMFF94 methods
 # 'GAUSSIAN' : FF methods supported by Gaussian (UFF, MMFF)
 # 'XTB' : GFN-FF method
 
@@ -37,7 +37,7 @@ DEFAULT_FF_LEVELS = {
     ### THESE TWO LINES
     'GAUSSIAN':'UFF',
     'XTB':'GFN-FF',
-    'OB':'UFF',
+    'OB':'MMFF94',
 }
 # Default levels used to run calculations, overridden by FFLEVEL keyword
 
