@@ -2,7 +2,7 @@
 '''
 
 TSCODE: Transition State Conformational Docker
-Copyright (C) 2021 Nicolò Tampellini
+Copyright (C) 2021-2022 Nicolò Tampellini
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -23,27 +23,27 @@ def run_tests():
     from subprocess import CalledProcessError
 
     os.chdir(os.path.dirname(os.path.realpath(__file__)))
-    # sys.path.append(os.getcwd())
 
-    from tscode.settings import COMMANDS, FF_OPT_BOOL, FF_CALC, CALCULATOR, PROCS, DEFAULT_LEVELS, DEFAULT_FF_LEVELS
+    from tscode.settings import (CALCULATOR, COMMANDS, DEFAULT_FF_LEVELS,
+                                 DEFAULT_LEVELS, FF_CALC, FF_OPT_BOOL, PROCS)
 
     if CALCULATOR not in ('MOPAC','ORCA','GAUSSIAN','XTB'):
         raise Exception(f'{CALCULATOR} is not a valid calculator. Use MOPAC, ORCA, GAUSSIAN or XTB.')
 
-    from tscode.optimization_methods import opt_funcs_dict
-    from tscode.utils import HiddenPrints, time_to_string, clean_directory, run_command, loadbar
-    from tscode.ase_manipulations import get_ase_calc
-    from cclib.io import ccread
-    from ase.optimize import LBFGS
-    from ase.atoms import Atoms
     import numpy as np
+    from ase.atoms import Atoms
+    from ase.optimize import LBFGS
+
+    from tscode.ase_manipulations import get_ase_calc
+    from tscode.optimization_methods import opt_funcs_dict
+    from tscode.utils import (HiddenPrints, clean_directory, loadbar, read_xyz,
+                              run_command, time_to_string)
 
-    os.chdir(os.path.dirname(os.getcwd()))
     os.chdir('tests')
 
-    t_start_run = time.time()
+    t_start_run = time.perf_counter()
 
-    data = ccread('C2H4.xyz')
+    data = read_xyz('C2H4.xyz')
 
     ##########################################################################
 
@@ -96,15 +96,16 @@ def run_tests():
                                     procs=PROCS,
                                     read_output=False)
 
-        print(f'{FF_CALC} raw calculator works.')
+        print(f'{FF_CALC} FF raw calculator works.')
 
         ##########################################################################
+        
+        if FF_CALC != 'OB':
+            atoms.calc = get_ase_calc((FF_CALC, DEFAULT_FF_LEVELS[FF_CALC], PROCS, None))
+            LBFGS(atoms, logfile=None).run()
 
-        atoms.calc = get_ase_calc((FF_CALC, DEFAULT_FF_LEVELS[FF_CALC], PROCS, None))
-        LBFGS(atoms, logfile=None).run()
-
-        clean_directory()
-        print(f'{FF_CALC} ASE calculator works.')
+            clean_directory()
+            print(f'{FF_CALC} ASE calculator works.')
 
     print('\nNo installation faults detected with the current settings. Running tests.')
 
@@ -112,7 +113,7 @@ def run_tests():
 
     tests = []
     for f in os.listdir():
-        if f.endswith('.txt'):
+        if f.endswith('.inp'):
             tests.append(os.path.realpath(f))
 
     # os.chdir(os.path.dirname(os.getcwd()))
@@ -124,16 +125,17 @@ def run_tests():
         name = f.split('\\')[-1].split('/')[-1][:-4] # trying to make it work for either Win, Linux (and Mac?)
         loadbar(i, len(tests), f'Running TSCoDe tests ({name}): ')
         
-        t_start = time.time()
+        t_start = time.perf_counter()
         try:
             with HiddenPrints():
                 run_command(f'python -m tscode {f} -n {name}')
 
         except CalledProcessError as error:
+            print('\n\n--> An error occurred:\n')
             print(error.stderr.decode("utf-8"))
             quit()
                     
-        t_end = time.time()
+        t_end = time.perf_counter()
         times.append(t_end-t_start)
 
     loadbar(len(tests), len(tests), f'Running TSCoDe tests ({name}): ')    
@@ -142,4 +144,4 @@ def run_tests():
     for i, f in enumerate(tests):
         print('    {:25s}{} s'.format(f.split('\\')[-1].split('/')[-1][:-4], round(times[i], 3)))
 
-    print(f'\nTSCoDe tests completed with no errors. ({time_to_string(time.time() - t_start_run)})\n')
\ No newline at end of file
+    print(f'\nTSCoDe tests completed with no errors. ({time_to_string(time.perf_counter() - t_start_run)})\n')
\ No newline at end of file
